/*
meparse.c
Mathematical expression parser
 
(C) 2002, Darren Wilkinson
d.j.wilkinson@ncl.ac.uk
http://www.staff.ncl.ac.uk/d.j.wilkinson/

Last edited on: $LastChangedDate: 2006-01-25 13:10:57 +0000 (Wed, 25 Jan 2006) $
by $Author: colin $
Revision: $Rev: 54 $
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <time.h>
#include "meparse.h"


/* internal function prototypes */
MEPNode * MEPParse(char *formula,void * envPtr,int type,gsl_vector *localParArray[]);
MEPNode * MEPMkNode(MEPType type);
char * MEPScan(char *formula,char op);
MEPNode * MEPMkInOp(char *formula,double (*op)(double,double),
		    char *p,void *envPtr,int type,gsl_vector *localParArray[]);
MEPNode * MEPMkPreOp(char *formula,double (*op)(double),
		     void *envPtr,int type,gsl_vector *localParArray[]);
MEPNode * MEPMkVar(char * formula,void *envPtr,int type,gsl_vector *localParArray[]);


/* binary infix functions */
double MEPInPlus(double arg1,double arg2);
double MEPInMinus(double arg1,double arg2);
double MEPInTimes(double arg1,double arg2);
double MEPInDivide(double arg1,double arg2);

/* binary operators */
double MEPGt(double arg1,double arg2);
double MEPLt(double arg1,double arg2);
double MEPGeq(double arg1,double arg2);
double MEPLeq(double arg1,double arg2);
double MEPEq(double arg1,double arg2);
double MEPNeq(double arg1,double arg2);
double MEPAnd(double arg1,double arg2);
double MEPOr(double arg1,double arg2);
double MEPUniformRandom(double arg1,double arg2);
double MEPUniformDiscreteRandom(double arg1,double arg2);
double MEPNormalRandom(double arg1,double arg2);
double MEPGammaRandom(double arg1,double arg2);

/* additional prefix function prototypes */
double MEPMinus(double arg);
double MEPAbs(double arg);
double MEPSqr(double arg);
double MEPExponentialRandom(double arg);
double MEPPoissonRandom(double arg);

/* infix operators */
typedef struct
{
  char chr;
  double (*ptr)(double,double);
} MEPInOp;

MEPInOp MEPInOpArray[] = {
  {'+',MEPInPlus},
  {'-',MEPInMinus},
  {'*',MEPInTimes},
  {'/',MEPInDivide},
  {'^',pow},
  {'>',MEPGt},
  {'<',MEPLt},
  {':',NULL}
};

/* unary prefix operators */
typedef struct
{
  char * string;
  char * match_string;
  double (*ptr)(double);
} MEPPreOp;

MEPPreOp MEPPreOpArray[] = {
  {"abs",   "abs(",   MEPAbs  },
  {"acos",  "acos(",  acos    },
  {"asin",  "asin(",  asin    },
  {"atan",  "atan(",  atan    },
  {"ceil",  "ceil(",  ceil    },
  {"cos",   "cos(",   cos     },
  {"exp",   "exp(",   exp     },
  {"floor", "floor(", floor   },
  {"log",   "log(",   log     },
  {"log10", "log10(", log10   },
  {"sin",   "sin(",   sin     },
  {"sqr",   "sqr(",   MEPSqr  },
  {"sqrt",  "sqrt(",  sqrt    },
  {"tan",   "tan(",   tan     },
  {"-",     "-(",     MEPMinus},
  {"exponentialRandom",    "exponentialRandom(",    MEPExponentialRandom },
  {"poissonRandom",	   "poissonRandom(", 	    MEPPoissonRandom     },
  {NULL,    NULL,     NULL    }
};

/* binary prefix operators */
typedef struct
{
  char * string;
  char * match_string;
  double (*ptr)(double,double);
} MEPBinOp;

MEPBinOp MEPBinOpArray[] = {
  {"pow",  "pow(", pow    },
  {"gt",   "gt(",  MEPGt  },
  {"lt",   "lt(",  MEPLt  },
  {"geq",   "geq(",  MEPGeq  },
  {"leq",   "leq(",  MEPLeq  },
  {"eq",   "eq(",  MEPEq  },
  {"neq",   "neq(",  MEPNeq  },
  {"and",   "and(",  MEPAnd  },
  {"or",   "or(",  MEPOr  },
  {"uniformRandom",   "uniformRandom(",   MEPUniformRandom   },
  {"uniformDiscreteRandom",   "uniformDiscreteRandom(",   MEPUniformDiscreteRandom   },
  {"normalRandom",    "normalRandom(",    MEPNormalRandom   },
  {"gammaRandom",    "gammaRandom(",    MEPGammaRandom   },
  {NULL,   NULL,   NULL   }
};


/*function definitions */
MEPNode * MEPParseTree(const char *formula,void * envPtr,int type,gsl_vector *localParArray[])
{
  /* first strip out white-space, and then call MEPParse */
  MEPNode * root;
  char *stripped,*sp;
  const char *fp;
  stripped=malloc(strlen(formula)*sizeof(char));
  sp=stripped;
  fp=formula;
  while (*fp) {
    if (!isspace(*fp)) {
      *sp=*fp;
      sp++;
    }
    fp++;
  }
  *sp='\0';
  root=MEPParse(stripped,envPtr,type,localParArray);
  free(stripped);
  return(root);
}

MEPNode * MEPParse(char *formula,void * envPtr,int type,gsl_vector *localParArray[])
{
  MEPNode *root;
  char *p;
  MEPInOp *in;
  MEPPreOp *pre;
  MEPBinOp *bin;
  /* uncomment the following line for parsing information */
 /* fprintf(stderr,"Parsing : %s\n",formula);  */      
  
  /* first check for infix operators */
  in=MEPInOpArray;
  while (in->ptr) {
    if ( (p=MEPScan(formula,in->chr)) ) {
      root=MEPMkInOp(formula,in->ptr,p,envPtr,type,localParArray);
      return(root);
    }
    in++;
  }
  /* now check for parentheses */
  if ( *formula == '(' ) {
    int len=strlen(formula);
    if ( !(formula[len-1] == ')') ) {
      fprintf(stderr,"Unmatched parentheses!\n");
      return(NULL);
    }
    formula[len-1]='\0';
    return(MEPParse(++formula,envPtr,type,localParArray));
  }
  /* now check for prefix operators (which must be unary) */
  pre=MEPPreOpArray;
  while (pre->ptr) {
    if ( strstr(formula,pre->match_string) == formula ) {
      root=MEPMkPreOp(formula,pre->ptr,envPtr,type,localParArray);
      return(root);
    }
    pre++;
  }
  /* now check for binary prefix operators */
  bin=MEPBinOpArray;
  while (bin->ptr) {
    if ( strstr(formula,bin->match_string) == formula ) {
      int len=strlen(formula);
      if (!(formula[len-1] == ')') ) {
	fprintf(stderr,"Unmatched parentheses in binary op!\n");
	return(NULL);
      }
      formula[len-1]='\0';
      formula=strchr(formula,'(');
      formula++;
      if ( !(p=MEPScan(formula,',')) ) {
	fprintf(stderr,"No comma in binary operator!\n");
      }
      root=MEPMkInOp(formula,bin->ptr,p,envPtr,type,localParArray);
      return(root);
    }
    bin++;
  }
  /* now check for unary minus */
  if ( *formula == '-' ) {
    root=MEPMkNode(MEPPrefix);
    (root->info).pre_op=MEPMinus;
    root->left=MEPParse(++formula,envPtr,type,localParArray);
    if (root->left == NULL) {
      return(NULL);
    }
    else {
      return(root);
    } 
  }
  /* now check for unary plus (does anyone use this?!) */
  else if ( *formula == '+') {
    return( MEPParse(++formula,envPtr,type,localParArray) );
  }
  /* now check for a variable */
  else if ( !isdigit(*formula) ) {
    root=MEPMkVar(formula,envPtr,type,localParArray);
    return(root);
  }
  /* now must be a constant */
  else {
    root=MEPMkNode(MEPConstant);
    (root->info).data = atof(formula);
    return(root);
  }
}

MEPNode * MEPMkInOp(char *formula,double (*op)(double,double)
		    ,char *p,void *envPtr,int type,gsl_vector *localParArray[])
{
  MEPNode *root;
  *p='\0';
  root=MEPMkNode(MEPInfix);
  (root->info).in_op=op ;
  root->left=MEPParse(formula,envPtr,type,localParArray);
  root->right=MEPParse(++p,envPtr,type,localParArray);
  if ( (root->left == NULL) || (root->right == NULL) ) {
    return(NULL);
  }
  else {
    return(root);
  }
}

MEPNode * MEPMkPreOp(char *formula,double (*op)(double),
		     void *envPtr,int type,gsl_vector *localParArray[])
{
  MEPNode *root;
  char *p;
  p=strchr(formula,'(');
  root=MEPMkNode(MEPPrefix);
  (root->info).pre_op=op;
  root->left=MEPParse(p,envPtr,type,localParArray);
  if (root->left == NULL) {
    return(NULL);
  }
  else {
    return(root);
  }
}

MEPNode * MEPMkVar(char * formula,void *envPtr,int type,gsl_vector *localParArray[])
{
  MEPNode *node;
  node=MEPMkNode(MEPVariable);
  MEPMkVarPtr(formula,node,envPtr,type,localParArray);
  if ( node->type == MEPVarERROR ) {
    return(NULL);
  }
  else {
    return(node);
  }
}

MEPNode * MEPMkNode(MEPType type)
{
  MEPNode *node;
  node=malloc(sizeof(MEPNode));
  node->type=type;
  return(node);
}

char * MEPScan(char *formula,char op)
{
  int level;
  char *p;
  level=0;
  p=formula;
  while (*p) {
    if (*p == '(') {
      level++;
    }
    else if (*p == ')') {
      level--;
    }
    else if ( (*p == op) && (level == 0) && (p != formula) ) {
      return(p);
    }
    p++;
  }
  return(NULL);
}

void MEPFreeTree(MEPNode *tree)
{
  switch(tree->type) {
  case MEPInfix: {
    MEPFreeTree(tree->left);
    MEPFreeTree(tree->right);
    free(tree);
    break;
  }
  case MEPPrefix: {
    MEPFreeTree(tree->left);
    free(tree);
    break;
  }
  default: {
    free(tree);
    break;
  }
  }
}

void MEPPrint(MEPNode *tree)
{
  MEPPreOp * pre;
  MEPInOp * in;
  MEPBinOp * bin;
  switch(tree->type) {
  case MEPInfix: {
    in=MEPInOpArray;
    while (in->ptr) {
      if ( (in->ptr) == (tree->info).in_op ) {
	printf("(%c ",in->chr);
	MEPPrint(tree->left);
	printf(" ");
	MEPPrint(tree->right);
	printf(")\n");
	return;
      }
      in++;
    }
    bin=MEPBinOpArray;
    while (bin->ptr) {
      if ( (bin->ptr) == (tree->info).in_op ) {
	printf("(%s ",bin->string);
	MEPPrint(tree->left);
	printf(" ");
	MEPPrint(tree->right);
	printf(")\n");
	return;
      }
      bin++;
    }
  }
  case MEPPrefix: {
    pre=MEPPreOpArray;
    while (pre->ptr) {
      if ( (pre->ptr) == (tree->info).pre_op ) {
	printf("(%s ",pre->string);
	MEPPrint(tree->left);
	printf(") ");
	return;
      }
      pre++;
    }
  }
  case MEPConstant: {
    printf("%f",(tree->info).data);
    return;
  }
  case MEPVariable: {
    printf("v=%f",*((tree->info).var));
    return;
  }
  case MEPVariableI: {
    printf("vI=%d",*((tree->info).varI));
    return;
  }
  case MEPVariableL: {
    printf("vL=%ld",*((tree->info).varL));
    return;
  }
  case MEPVariableF: {
    printf("vF=%f",*((tree->info).varF));
    return;
  }
  default: {
    fprintf(stderr,"Unknown node type: %d\n",tree->type);
    exit(EXIT_FAILURE);
  }
  }
}

double MEPEval(MEPNode *tree)
{
  switch(tree->type) {
  case MEPInfix: {
    return(((tree->info).in_op)(MEPEval(tree->left),MEPEval(tree->right)));
  }
  case MEPPrefix: {
    return( ((tree->info).pre_op)(MEPEval(tree->left)) );
  }
  case MEPConstant: {
    return( (tree->info).data );
  }
  case MEPVariable: {
    return(*((tree->info).var) );
  }
  case MEPVariableI: {
    return( (double) *((tree->info).varI) );
  }
  case MEPVariableL: {
    return( (double) *((tree->info).varL) );
  }
  case MEPVariableF: {
    return( (double) *((tree->info).varF) );
  }
  default: {
    fprintf(stderr,"Unknown node type.\n");
    exit(EXIT_FAILURE);
  }
  }
}

/* binary infix functions */
double MEPInPlus(double arg1,double arg2) {
  return( arg1 + arg2 ); }

double MEPInMinus(double arg1,double arg2) {
  return( arg1 - arg2 ); }

double MEPInTimes(double arg1,double arg2) {
  return( arg1 * arg2 ); }

double MEPInDivide(double arg1,double arg2) {
  return( arg1 / arg2 ); }

/* binary operators */
double MEPGt(double arg1,double arg2) {
  return( arg1 > arg2 ); }

double MEPLt(double arg1,double arg2) {
  return( arg1 < arg2 ); }

double MEPGeq(double arg1,double arg2) {
  return( arg1 >= arg2 ); }

double MEPLeq(double arg1,double arg2) {
  return( arg1 <= arg2 ); }

double MEPEq(double arg1,double arg2) {
  return( arg1 == arg2 ); }

double MEPNeq(double arg1,double arg2) {
  return( arg1 != arg2 ); }

double MEPAnd(double arg1,double arg2) {
  return( arg1 && arg2 ); }

double MEPOr(double arg1,double arg2) {
  return( arg1 || arg2 ); }                                                              


double MEPUniformRandom(double arg1, double arg2){
  gsl_rng *r;
  r=get_globalr();       
  return(gsl_ran_flat(r,arg1,arg2)); 
  }     

double MEPUniformDiscreteRandom(double arg1, double arg2){
  gsl_rng *r;
  r=get_globalr();       
  return(floor(gsl_ran_flat(r,arg1,arg2))); 
  }     


double MEPNormalRandom(double arg1, double arg2){
  gsl_rng *r;
  r=get_globalr();       
  return(arg1+gsl_ran_gaussian(r,arg2)); 
  }     

double MEPGammaRandom(double arg1, double arg2){
  gsl_rng *r;
  r=get_globalr();       
  return(gsl_ran_gamma(r,arg1,arg2)); 
  }     


/* additional prefix functions */
double MEPMinus(double arg) {
  return( -arg ); }

double MEPAbs(double arg) {
  if ( arg < 0 ) {
    return(-arg);
  }
  else {
    return(arg);
  }
}

double MEPSqr(double arg) {
  return( arg * arg ); }


double MEPExponentialRandom(double arg){
  gsl_rng *r;
  r=get_globalr();       
  return(gsl_ran_exponential(r,arg)); 
  }     


double MEPPoissonRandom(double arg){
  gsl_rng *r;
  r=get_globalr();       
  return(1.0*gsl_ran_poisson(r,arg)); 
  }  

/* eof */

