/*
simulate.c

Gillespie algorithm simulation function
together with associated helper functions
amended to include MathML and events.

Last edited on: $LastChangedDate: 2006-01-25 13:10:57 +0000 (Wed, 25 Jan 2006) $
by $Author: colin $
Revision: $Rev: 54 $

*/


#include "simulate.h"

/* private function prototypes */
void setTypes(void);
void setRates(void);
void outputHeader(FILE *);
void outputCurrentState(FILE *);
double vector_sum(gsl_vector *);
int ran_discrete(gsl_rng *,gsl_vector *);
void updateState(int);
MEPNode ** parseFormulas(gsl_vector *localParArray[]);
MEPNode ** parseTriggers(gsl_vector *localParArray[]);
MEPNode ** parseRules(gsl_vector *assignmentRulesArray[]);
void MEPMkVarPtr(char *varStr,MEPNode *varNode,void *envPtr,int type,gsl_vector *localParArray[]);
void  do_events(gsl_vector *evArray,gsl_vector *localParArray[]);
void  initialise_events(gsl_vector *evArray,gsl_vector *localParArray[]);
void do_assignments(const char *varStr, double asval);
void set_globalParVec(void);
void set_localParVec(gsl_vector *localParArray[]);
void set_spVec(void);
void do_rules(void);
void checkBoundaryConditions(Model_t *mod);
int check_events(gsl_vector *evArray);
gsl_rng * set_globalr(gsl_rng *r);

/* file global variables */
Model_t *mod;
double terminalTime,simtime,output_time;
int numReactions,numSpecies,numGlobalParameters,numLocalParameters,numRules,numTimedEvents;
long counter;
unsigned int  numEvents;
char** outputArray;
int outputArrayLen;
MEPNode **reacArray;
MEPNode **trigArray;
MEPNode **rulesArray;
gsl_vector *reacRates;
gsl_vector *globalParVec;
gsl_vector *spVec;
gsl_vector **localParArray;
gsl_vector **assignmentRulesArray;
gsl_vector *evArray;
gsl_rng *globalr;

/* main simulation function */
void simulate(Model_t *m,FILE *s,gsl_rng *r,double tt,int fileref,int numSteps,char** explodedarray,int arraysize)
{
  double combinedRate,delay,newdelay,time_step;
  int i,j,reacInd,outputReaction;
  Reaction_t *reac;/* CHANGE:  Added reaction data to find what species are reacting. */
  
  /* initialisations */
  outputArray = explodedarray;
  outputArrayLen = arraysize;
  
  mod=m;
  checkBoundaryConditions(mod); /*set stoichiometry of any species with boundarycondition==true to zero */
  terminalTime=tt;
  simtime=0;
  output_time=0;
  time_step=terminalTime/(1.0*numSteps);
  numReactions=Model_getNumReactions(mod);
  numSpecies=Model_getNumSpecies(mod);
  numRules=Model_getNumRules(mod);
  numEvents=Model_getNumEvents(mod);
  numGlobalParameters=Model_getNumParameters(mod);
  reacRates=gsl_vector_alloc(numReactions);
  localParArray=malloc(numReactions*sizeof(gsl_vector *));
  set_localParVec(localParArray);    /*initialises array for local parameter values */
  if(numGlobalParameters>0){
      globalParVec=gsl_vector_alloc(numGlobalParameters);
      set_globalParVec();       /*initialises array for global parameter values */
  }
  spVec= gsl_vector_alloc(numSpecies);
  set_spVec();         /*initialsies array for species amounts */
  if(numRules>0)
    assignmentRulesArray=malloc(numRules*sizeof(gsl_vector *));
  if(numEvents>0)
     evArray=gsl_vector_alloc(numEvents);
  for(i=0;i<numEvents;i++)
       gsl_vector_set(evArray,i,0); /*initialise evArray - all triggers are false at start*/
  reacArray=parseFormulas(localParArray);
  trigArray=parseTriggers(localParArray);
  rulesArray=parseRules(assignmentRulesArray);
  do_rules();
  initialise_events(evArray,localParArray);
  globalr=set_globalr(r); /*for use in meparse.c */
  outputHeader(s);            /*output name of species to file */
  if(fileref>0) dbUpdate(fileref,output_time,mod,spVec);
  else  outputCurrentState(s);     /* output state at time t=0 */
  output_time+=time_step;
  globalr=set_globalr(r);  /*for use in meparse.c */
 
 /* main simulation loop */
  while (simtime < terminalTime) {
    do_rules();
    do_events(evArray,localParArray);
    setRates();    /* sets the vector "reacRates" */
    combinedRate=vector_sum(reacRates);
    if(combinedRate>100000000000.0){
      fprintf(stderr, "Combined reaction rate is too high, check initial conditions and parameter values");
      exit(EXIT_FAILURE);
    }          
    delay=gsl_ran_exponential(r,1/combinedRate); /* parameterised by mean!  */
    simtime+=delay;
    newdelay=check_events(evArray);
    while(newdelay!=0){
       simtime=newdelay;
       while(output_time<simtime&&output_time<=terminalTime){
           if(fileref>0) dbUpdate(fileref,output_time,mod,spVec);
           else outputCurrentState(s);
           output_time+=time_step;
       }
       do_events(evArray,localParArray);
       setRates();   /* sets the vector "reacRates" */
       combinedRate=vector_sum(reacRates); 
       delay=gsl_ran_exponential(r,1/combinedRate); /* parameterised by mean!  */
       simtime+=delay;
       newdelay=check_events(evArray);                        
    }
    /* CHANGE:  Outputs every molecular event now. */
    if (simtime < terminalTime) {
      reacInd=ran_discrete(r,reacRates);     /* choose a reaction index */
      
      
      outputReaction = 0;
      
      reac=Model_getReaction(mod,reacInd);
      /* Search all species involved in the reaction and if one is important, then output this
      reaction. */
      for (i=0;i<Reaction_getNumReactants(reac);i++) {


        for (j=0;j<outputArrayLen;j++) {
          if (!strcmp(Species_getId(Model_getSpeciesById(mod,SpeciesReference_getSpecies(Reaction_getReactant(reac,i)))), outputArray[j])) {
            outputReaction = 1;
	    break;
          }
	}
      }
      
      for (i=0;i<Reaction_getNumProducts(reac);i++) {
        for (j=0;j<outputArrayLen;j++) {
          if (!strcmp(Species_getId(Model_getSpeciesById(mod,SpeciesReference_getSpecies(Reaction_getProduct(reac,i)))), outputArray[j])) {
            outputReaction = 1;
	    break;
          }
	}
      }
      
      updateState(reacInd); /* update the state according to chosen reaction */
      if (outputReaction) {
        outputCurrentState(s); 
      }
    }
    counter++;
  }
  gsl_vector_free(reacRates);
  gsl_vector_free(spVec);
  if(numGlobalParameters>0) gsl_vector_free(globalParVec);

  if (numLocalParameters>0) {
      for (i=0;i<numLocalParameters;i++) {
        gsl_vector_free(localParArray[i]);
      }
      free(localParArray);
  }
/*  if (numRules>0) {
    for (i=0;i<numRules;i++) {
       gsl_vector_free(assignmentRulesArray[i]);
    }
  free(assignmentRulesArray);
  }*/
  if(numEvents>0) gsl_vector_free(evArray);
  /* CHANGE:  Added a new line after every simulator. */
  fprintf(s,"\n");

}

/* helper functions */
MEPNode ** parseFormulas(gsl_vector *localParArray[])
{
  int i,type=1;
  const char *s;
  Reaction_t *reac;
  KineticLaw_t *kl;
  MEPNode **array;
  const ASTNode_t *node;
  array=malloc(numReactions*sizeof(MEPNode *));
  if (!array) {
    perror("failed malloc");
    exit(EXIT_FAILURE);
  }
  for (i=0;i<numReactions;i++) {
    reac=Model_getReaction(mod,i);
    kl=Reaction_getKineticLaw(reac);
    if(KineticLaw_isSetFormula(kl))
        s=KineticLaw_getFormula(kl);
    else{
        node=KineticLaw_getMath(kl);
        s=SBML_formulaToString(node);
    }
    array[i]=MEPParseTree(s,reac,type,localParArray);
    /* MEPPrint(array[i]);   */
    if (!array[i]) {
      fprintf(stderr,"Error parsing kinetic law: %s\n",s);
      exit(EXIT_FAILURE);
    }
  }
  return(array);
}

MEPNode ** parseTriggers(gsl_vector *localParArray[])
{
  int i,type;
  const char *s;
  Event_t *event;
  MEPNode **array;
  const ASTNode_t *tr;
  array=malloc(numEvents*sizeof(MEPNode *));
  if (!array) {
    perror("failed malloc");
    exit(EXIT_FAILURE);
  }
  type=2; /* for parsing trigger */
  for(i=0;i<numEvents;i++){
    event=Model_getEvent(mod,i);
    tr=Event_getTrigger(event);
    s=SBML_formulaToString(tr);
    array[i]=MEPParseTree(s,event,type,localParArray);
   /* MEPPrint(array[i]);    */     
    if (!array[i]) {
      fprintf(stderr,"Error parsing triggers\n");
      exit(EXIT_FAILURE);
    }
  }
  return(array);
}

MEPNode ** parseRules(gsl_vector *assignmentRulesArray[])
{
  int i,type=2; /* local parameters are not allowed in rules */
  const char *s;
  Rule_t *r;
  MEPNode **array;
  const ASTNode_t *node;
  array=malloc(numRules*sizeof(MEPNode *));
  if (!array) {
    perror("failed malloc");
    exit(EXIT_FAILURE);
  }
  for (i=0;i<numRules;i++){
    r=Model_getRule(mod,i);
    if(SBase_getTypeCode((SBase_t *)r)!=SBML_ASSIGNMENT_RULE){
       fprintf(stderr, "This simulator only accepts rules of type assignment rule. \n");
       exit(EXIT_FAILURE);
    }
    if(Rule_isSetFormula(r))
        s=Rule_getFormula(r);
    else{
        node=Rule_getMath(r);
        s=SBML_formulaToString(node);
    }

    array[i]=MEPParseTree(s,r,type,assignmentRulesArray);
    if (!array[i]) {
      fprintf(stderr,"Error parsing kinetic law: %s\n",s);
      exit(EXIT_FAILURE);
    }
  }
  return(array);
}

void MEPMkVarPtr(char *varStr,MEPNode *varNode,void *envPtr,int type, gsl_vector *localParArray[])
{
  /* MEP callback function */
  Reaction_t *reac, *r;
  KineticLaw_t *kl;
  Parameter_t *par;
  Species_t *spec;
  int i,j,reacNo;
  reacNo=-1;
  
   /* first check for local parameters */
  if(type==1){
    reac=(Reaction_t *) envPtr;
    for(j=0;j<numReactions;j++){
       r=Model_getReaction(mod,j);
       if(!strcmp(Reaction_getId(r),Reaction_getId(reac))) reacNo=j;
    }
    kl=Reaction_getKineticLaw(reac);
    for(i=0;i<KineticLaw_getNumParameters(kl);i++){          
          par=KineticLaw_getParameter(kl,i);
          if (!strcmp(varStr,Parameter_getId(par))) {
	        if(reacNo!=-1){
                   (varNode->info).var=gsl_vector_ptr(localParArray[reacNo],i);
                   return;
		}
		else{
		   fprintf(stderr,"Error checking local parameters");
		   exit(EXIT_FAILURE);
		}
          }
     }
  } 
  /* next check for global parameters if any exist */
  if(numGlobalParameters>0){
   for(i=0;i<Model_getNumParameters(mod);i++){
          par=Model_getParameter(mod,i);
          if (!strcmp(varStr,Parameter_getId(par))){
             (varNode->info).var=gsl_vector_ptr(globalParVec,i);
             return;                  
         }
    }
  }
  /*next check for time variable */
  if(!strcmp(varStr,"t")){
          (varNode->info).var=&simtime;
          if(type==2) numTimedEvents++;
          return;
  }    
  /* next check for species */
  for (i=0;i<numSpecies;i++) {
    spec=Model_getSpecies(mod,i);
    if (!strcmp(varStr,Species_getId(spec))) {
              (varNode->info).var=gsl_vector_ptr(spVec,i);
               return;
     }
  }
  /* finally check for compartments */

  /* ######### not yet implemented ########### */

  /* failed to find a match - return an error... */
  fprintf(stderr,"Variable %s not found\n",varStr);
  varNode->type = MEPVarERROR;
  return;
}

void updateState(int ind)
{
  int i,j;
  Reaction_t *reac;
  SpeciesReference_t *sr;
  Species_t *sp, *spec;
  double st;
  reac=Model_getReaction(mod,ind);
  for (i=0;i<Reaction_getNumReactants(reac);i++) {        /*this loop checks that no species can go negative during reaction*/
    sr=Reaction_getReactant(reac,i);
    st=SpeciesReference_getStoichiometry (sr);
    sp=Model_getSpeciesById(mod,SpeciesReference_getSpecies(sr));
    for(j=0;j<numSpecies;j++){
        spec=Model_getSpecies(mod,j);
        if(!strcmp(Species_getId(spec),Species_getId(sp))){
             if(gsl_vector_get(spVec,j)-st <0){
                 fprintf(stderr,"Species %s has gone negative in reaction %d\n",Species_getId(spec),(ind+1));
                 exit(EXIT_FAILURE);
             }
        }
     }
  }  
  for (i=0;i<Reaction_getNumReactants(reac);i++) {
     sr=Reaction_getReactant(reac,i);
     sp=Model_getSpeciesById(mod,SpeciesReference_getSpecies(sr));
     for(j=0;j<numSpecies;j++){
        spec=Model_getSpecies(mod,j);
        if(!strcmp(Species_getId(spec),Species_getId(sp))) {
             gsl_vector_set(spVec,j,(gsl_vector_get(spVec,j)-SpeciesReference_getStoichiometry(sr)));
        }   
     }     
  }
  for (i=0;i<Reaction_getNumProducts(reac);i++) {
        sr=Reaction_getProduct(reac,i);
        sp=Model_getSpeciesById(mod,SpeciesReference_getSpecies(sr));
        for(j=0;j<numSpecies;j++){
           spec=Model_getSpecies(mod,j);
           if(!strcmp(Species_getId(spec),Species_getId(sp)))
             gsl_vector_set(spVec,j,(gsl_vector_get(spVec,j)+SpeciesReference_getStoichiometry(sr)));
       }   
  }
}


void setRates(void)
{
  int i;
  MEPNode *pt;
  for (i=0;i<numReactions;i++) {
    pt=reacArray[i];
    gsl_vector_set(reacRates,i, MEPEval(pt) );
  }
}

void set_globalParVec(void)
{
  int i;
  Parameter_t *par;
  for(i=0;i<numGlobalParameters;i++){
      par=Model_getParameter(mod,i);
      gsl_vector_set(globalParVec,i,Parameter_getValue(par));
  }
}

void set_localParVec(gsl_vector *localParArray[])
{
  int i,j,numLocalParameters;
  Parameter_t *par;
  Reaction_t *reac;
  KineticLaw_t *kl;
  for(i=0;i<numReactions;i++){          /*get  local parameters */
      reac=Model_getReaction(mod, i);
      if(Reaction_isSetKineticLaw(reac)!=1){
         fprintf(stderr,"Error: Kinetic law is missing from reaction %d\n", i+1);
         exit(EXIT_FAILURE);
      }  
      kl=Reaction_getKineticLaw(reac);
      numLocalParameters=KineticLaw_getNumParameters(kl);
      if(numLocalParameters>0) localParArray[i]=gsl_vector_alloc(numLocalParameters);
      if(numLocalParameters>0){
        for(j=0;j<numLocalParameters;j++){
          par=KineticLaw_getParameter(kl,j);
          gsl_vector_set(localParArray[i],j,Parameter_getValue(par));
        }
      }
  }
  
}
        
void set_spVec(void)
{
 int i;
 Species_t *sp;
 for(i=0;i<numSpecies;i++){
      sp=Model_getSpecies(mod,i);
      gsl_vector_set(spVec,i,Species_getInitialAmount(sp));
 }
}

void do_rules(void)
{                       /*assignment rules  only */
 int i,j;
 double value;
 const char *s;
 Rule_t *r;
 Species_t *sp;
 AssignmentRule_t *ar;
 MEPNode *pt;
 for ( i=0; i<numRules; i++ ) {
      r = Model_getRule(mod,i);
      if (SBase_getTypeCode((SBase_t *) r)  == SBML_ASSIGNMENT_RULE ) {
          ar=(AssignmentRule_t *)r;
          s=AssignmentRule_getVariable(ar);
          pt=rulesArray[i];
          value=MEPEval(pt);
          for(j=0;j<numSpecies;j++){
              sp=Model_getSpecies(mod,j);
              if(!strcmp(Species_getId(sp),s)){
                gsl_vector_set(spVec,j,value);
              }
          }
         
      }
 }
}  

void outputHeader(FILE * s)
{
  int i,j;
  Species_t *sp;
  fprintf(s,"Time ");
  for (i=0;i<numSpecies;i++) {
    sp=Model_getSpecies(mod,i);
    
    for (j=0;j<outputArrayLen;j++) {
      if (!strcmp(Species_getId(sp), outputArray[j])) {
        fprintf(s,"%s ",Species_getId(sp));
	break;
      }
    }
  }
  fprintf(s,"\n");
}

void outputCurrentState(FILE * s)
{
  int i,j;
  Species_t *sp;
  /* CHANGE:  Changed to output the time the event happened rather than the current snap-shot time. */
  fprintf(s,"%f ",simtime);
  for (i=0;i<numSpecies;i++) {
    sp=Model_getSpecies(mod,i);
    
    for (j=0;j<outputArrayLen;j++) {
      if (!strcmp(Species_getId(sp), outputArray[j])) {
        fprintf(s,"%.0f ",gsl_vector_get(spVec,i));
        break;
      }
    }
  }
  fprintf(s,"\n");
}

int ran_discrete(gsl_rng * r,gsl_vector * v)
{
  int ind;
  double u,sum,runningSum;
  sum=vector_sum(v);
  runningSum=0;
  u=sum*gsl_ran_flat(r,0,1);
  for (ind=0;ind < (v->size);ind++) {
    runningSum+=gsl_vector_get(v,ind);
    if (runningSum > u) break;
  }
  if (ind == (v->size)) ind--;
  return(ind);
}

double vector_sum(gsl_vector * v)
{
  int i;
  double sum;
  sum=0;
  for (i=0;i<v->size;i++) {
    sum+=gsl_vector_get(v,i);
  }
  return(sum);
}

void do_events(gsl_vector *evArray,gsl_vector *localParArray[])
{
 int i,j;
 unsigned int numEvAss;
 Event_t *event;
 EventAssignment_t *evAss;
 ASTNode_t *tr;
 const ASTNode_t *as;
 MEPNode *trigger, *assignment;
 char *str;
 const char *eavar;
 double val,asval;
 tr=ASTNode_create();
 for(i=0;i<numEvents;i++){
     trigger=trigArray[i];
     val=MEPEval(trigger);        /*evaluate trigger */
     /* if(trigger was false at last iteration and is now true, trigger is fired  */
     if(val==1 && gsl_vector_get(evArray,i)==0){
         event=Model_getEvent(mod,i);
         numEvAss=Event_getNumEventAssignments(event);
         for(j=0;j<numEvAss;j++){
           evAss=Event_getEventAssignment(event,j);
           eavar=EventAssignment_getVariable(evAss);
           as=EventAssignment_getMath(evAss);
           str=SBML_formulaToString(as);
           assignment=MEPParseTree(str,event,2,localParArray);
           asval=MEPEval(assignment);
           do_assignments(eavar,asval);
           gsl_vector_set(evArray,i,1);
         }
         
     }
        /*if trigger was true and is now false */
     val=MEPEval(trigger);
     if(val==0 && gsl_vector_get(evArray,i)==1)
         gsl_vector_set(evArray,i,0);
      /*if trigger was  false and is still false OR if trigger was true and is still true go to next event */
 }
  ASTNode_free(tr);
}

void do_assignments(const char *varStr,double asval)
{
  Parameter_t *par;
  Species_t *spec;
  int i;
   /*first check for species */
  for(i=0;i<numSpecies;i++){
    spec=Model_getSpecies(mod,i);
    if(!strcmp(varStr,Species_getId(spec))){
       gsl_vector_set(spVec,i,asval);
       return;
    }
  }
    /*then check for global parameters */
  for(i=0;i<numGlobalParameters;i++){
      par=Model_getParameter(mod,i);
      if(!strcmp(varStr,Parameter_getId(par))){
        gsl_vector_set(globalParVec ,i,asval);
        return;
      }
  }
}

void initialise_events(gsl_vector *evArray,gsl_vector *localParArray[])
{
 int i;
 MEPNode *trigger;
 double val;
 for(i=0;i<numEvents;i++){
     trigger=trigArray[i];
     val=MEPEval(trigger);        /*evaluate trigger */

   /* if(trigger for event i is true at start of simulation set evArray[i]=1, so that event will
   not fire when events are checked for the first time */
       if(val==1 ){
          gsl_vector_set(evArray,i,1);
       }
                
 }
  
}

void checkBoundaryConditions(Model_t *mod)
{
  int numReactions,numSpecies,numReactants,numProducts;
  unsigned int i,j,k,l;
  Species_t *spec;
  Reaction_t *reac;
  SpeciesReference_t *sr;
  const char *name;
  numSpecies=Model_getNumSpecies(mod);
  numReactions=Model_getNumReactions(mod);
  for(i=0;i<numSpecies;i++){
       spec=Model_getSpecies(mod,i);
       if(Species_getBoundaryCondition(spec)){
             for(j=0;j<numReactions;j++){
                  reac=Model_getReaction(mod,j);
                  numReactants=Reaction_getNumReactants(reac);
                  numProducts=Reaction_getNumProducts(reac);
                  for(k=0;k<numReactants;k++){
                       sr=Reaction_getReactant(reac,k);
                       name=SpeciesReference_getSpecies(sr);
                       if(!strcmp(name,Species_getId(spec)))
                           SpeciesReference_setStoichiometry(sr,0.0);
                  }
                  for(l=0;l<numProducts;l++){
                       sr=Reaction_getProduct(reac,l);
                       name=SpeciesReference_getSpecies(sr);
                       if(!strcmp(name,Species_getId(spec)))
                         SpeciesReference_setStoichiometry(sr,0.0);
                  }
             }
       }
  }
}

int check_events(gsl_vector *evArray)
{
  int i;
  double val,newdelay;
  MEPNode *trigger;
  
  for(i=0;i<numEvents;i++){
    trigger=trigArray[i];
    val=MEPEval(trigger);
    if(val==1 && gsl_vector_get(evArray,i)==0){
    newdelay=MEPEval(trigger->right);
    return newdelay;
    }
  }
  newdelay=0;
  return newdelay;  
}

gsl_rng * set_globalr(gsl_rng *r){
return    r;
}    

gsl_rng * get_globalr(){
return globalr; 
}                                 

/* eof */
